"""
Small script that wraps an instance of CARTA, performing startup and shutdown procedures

:author: Nathan Hertz
"""
import argparse
import logging
import os
import random
import re
import signal
import socket
import string
import subprocess
import sys
from pathlib import Path
from types import FrameType
from typing import Dict, Optional, Tuple

import redis
from pycapo import CapoConfig

# Type aliases
RedisValues = Dict[str, str]
SetupValues = Tuple[argparse.Namespace, int, int, str]

_DESCRIPTION = (
    "Small script that wraps an instance of CARTA, performing startup and shutdown procedures"
)

# Logging
logger = logging.getLogger("carta_valet")
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler(sys.stdout))

# Assigned in main
REDIS_VALUES: RedisValues = {}
CONN: Optional[redis.Redis] = None
CARTA_PROCESS: Optional[subprocess.Popen] = None


def signal_handler(signum: int, frame: FrameType):
    """
    Handle a signal received by tearing down the program and exiting
    NOTE: Parameters signum and frame are convention for signal handlers, even if they don't end up used

    :param signum: Signal number
    :param frame: Stack frame
    """
    if signum == signal.SIGINT:
        logger.warning("WARNING: SIGINT received. Exiting.")
    elif signum == signal.SIGALRM:
        logger.warning("WARNING: Timeout elapsed. Exiting.")
    teardown_carta()
    sys.exit()


def generate_random_str() -> str:
    """
    Generates a random sequence of letters and digits of length 12

    :return: Random sequence of letters and digits of length 12
    """
    length = 12
    valid_chars = string.ascii_letters + string.digits
    return "".join(random.choice(valid_chars) for _ in range(length))


def get_available_port() -> int:
    """
    Asks the operating system for an available port on localhost using the socket module

    :return: Number of open port
    """
    with socket.socket(family=socket.AF_INET, type=socket.SOCK_STREAM) as s:
        # Bind socket to available port
        s.bind(("", 0))
        return s.getsockname()[1]


def connect_to_redis(host: str, port: int, password: str = None) -> redis.Redis:
    """
    Connect to Redis server located at host:port

    :param host: Host name
    :param port: Port
    :param password: Password for redis server; defaults to None
    :raises redis.ConnectionError: If the host or port are invalid
    :return: Successful connection to Redis server
    """
    conn = redis.Redis(host, port, password=password)
    # Ping the network to ensure the connection is sound
    conn.ping()

    return conn


def duplicate_redis_values_found(redis_values: RedisValues, conn: redis.Redis) -> bool:
    """
    Queries the Redis server to check if any of the values already exist

    :param redis_values: Dictionary of Redis entries
    :param conn: Connection to the Redis server
    :return: True if duplicates found; false if not
    """
    for key in redis_values:
        if conn.get(key):
            # Key is found
            return True

    return False


def get_redis_values(
    reverse_proxy_host: str,
    front_end_id: str,
    back_end_id: str,
    front_end_port: int,
    back_end_port: int,
) -> RedisValues:
    """
    Get Redis key/value pairs for CARTA front end and back end setup

    :param reverse_proxy_host: Hostname for the reverse proxy
    :param front_end_id: Random string ID for CARTA front end
    :param back_end_id: Random string ID for CARTA back end
    :param front_end_port: Port to run the CARTA front end with
    :param back_end_port: Port to run the CARTA back end with
    :return: Dictionary of values
    """
    service_name = generate_random_str()
    front_end = f"carta-front-{service_name}"
    back_end = f"carta-back-{service_name}"
    hostname = socket.getfqdn()
    values = {
        # FRONT END
        f"traefik/http/routers/{front_end}/rule": f"Host(`{reverse_proxy_host}`) && PathPrefix(`/{front_end_id}/`)",
        f"traefik/http/routers/{front_end}/service": front_end,
        f"traefik/http/services/{front_end}/loadbalancer/servers/0/url": f"http://{hostname}:{front_end_port}/",
        f"traefik/http/routers/{front_end}/middlewares/0": "stripPrefixFE@file",
        # BACK END
        f"traefik/http/routers/{back_end}/rule": f"Host(`{reverse_proxy_host}`) && PathPrefix(`/{back_end_id}/`)",
        f"traefik/http/routers/{back_end}/service": back_end,
        f"traefik/http/services/{back_end}/loadbalancer/servers/0/url": f"http://{hostname}:{back_end_port}/",
        f"traefik/http/routers/{back_end}/middlewares/0": "upgradeWSHeader@file",
        f"traefik/http/routers/{back_end}/middlewares/1": "stripPrefixFE@file",
    }
    return values


def set_redis_values(redis_values: RedisValues, timeout_minutes: int):
    """
    Set Redis values with an expiration of 24 hours

    :param redis_values: Dictionary of Redis entries
    :param timeout_minutes: Timeout of the valet in minutes
    """
    for key, val in redis_values.items():
        CONN.setex(key, 60 * timeout_minutes + 60, val)


# /home/local/Linux/rhel7/x86_64/carta-1.4.0/CARTA.AppImage


def setup_carta(redis_values: RedisValues, timeout_minutes: int):
    """
    Sets valet's Redis values and adds a signal handler in preparation for running CARTA

    :param redis_values: Dictionary of Redis entries
    :param timeout_minutes: Timeout of the valet in minutes
    """
    logger.info("Setting up...")
    # Set Redis values for front end and back end
    set_redis_values(redis_values, timeout_minutes)
    # Add handler for SIGINT so if the user ends the program using Ctrl+C, it'll close properly
    signal.signal(signal.SIGINT, signal_handler)


def run_carta(
    path_to_carta: str,
    timeout_minutes: int,
    file_browser_path: Path,
    front_end_port: int,
    back_end_port: int,
    carta_url: str,
):
    """
    Runs CARTA executable. Will close after a certain amount of time.

    :param path_to_carta: Path to CARTA executable
    :param timeout_minutes: Number of minutes to execute CARTA for before stopping it
    :param file_browser_path: Path to use for CARTA file browser
    :param front_end_port: Port to run the CARTA front end with
    :param back_end_port: Port to run the CARTA back end with
    :param carta_url: URL used to access CARTA from a web browser
    """
    global CARTA_PROCESS

    if not file_browser_path.exists():
        logger.error(
            f"ERROR: Invalid file browser path. {file_browser_path}: Directory does not exist."
        )
        return
    elif not file_browser_path.is_dir():
        logger.error(
            f"ERROR: Invalid file browser path. {file_browser_path}: Path is not a directory."
        )
        return

    logger.info(f"Running CARTA with a timeout of {timeout_minutes} minute(s)...")
    if re.search(r"CARTA.AppImage", path_to_carta):
        # Command uses AppImage, include --remote option
        carta_command = [
            path_to_carta,
            "--remote",
            f"--port={str(back_end_port)}",
            f"--fport={str(front_end_port)}",
            f"--folder={file_browser_path}",
            f"--root={file_browser_path}",
        ]
    else:
        carta_command = [
            path_to_carta,
            f"--port={str(back_end_port)}",
            f"--fport={str(front_end_port)}",
            f"--folder={file_browser_path}",
            f"--root={file_browser_path}",
        ]
    try:
        CARTA_PROCESS = subprocess.Popen(
            carta_command,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
    except (OSError, FileNotFoundError, subprocess.SubprocessError) as err:
        teardown_carta()
        sys.exit(f"ERROR: Failed to launch CARTA: {err}")
    else:
        # Activate timeout handler
        signal.signal(signal.SIGALRM, signal_handler)
        signal.alarm(60 * timeout_minutes)
        logger.info(f"To access CARTA, enter this URL into a web browser: {carta_url}")
        logger.info("Press Ctrl+C to quit.")

        # Loop until process ends or timeout elapses, whichever happens first
        while CARTA_PROCESS.returncode is None:
            CARTA_PROCESS.poll()


def teardown_carta():
    """
    Deletes instance's Redis keys from the Redis server
    """
    logger.info("Cleaning up...")
    for key in REDIS_VALUES:
        CONN.delete(key)
    if CARTA_PROCESS and CARTA_PROCESS.returncode is None:
        # If CARTA is running, kill it
        try:
            # End the process (simulate Ctrl+Z)
            CARTA_PROCESS.send_signal(signal.SIGSTOP)
            # Kill the process
            os.killpg(os.getpgid(CARTA_PROCESS.pid), signal.SIGTERM)
            CARTA_PROCESS.wait()
        except OSError:
            # Process isn't running; just pass
            pass
        else:
            logger.info("Killed running CARTA process.")
    else:
        logger.warning("WARNING: CARTA not running.")


def set_args(args: argparse.Namespace) -> argparse.Namespace:
    """
    Sets argument default values for any arguments not provided

    :param args: Args collected from the command line
    :return: Fully-completed args
    """
    if hasattr(args, "profile"):
        args.profile = CapoConfig(profile=args.profile)
    else:
        # No profile specified
        args.profile = CapoConfig()
    if not args.timeout:
        # No timeout set
        args.timeout = args.profile.get("edu.nrao.carta.timeout-minutes")
    if not args.where:
        # No where set
        args.where = Path.cwd()
    elif args.where[0] != "/":
        # Path is relative
        args.where = Path(Path.cwd() / args.where)
    else:
        # Where set via command line
        args.where = Path(args.where)
    args.timeout = int(args.timeout)
    args.path_to_carta = args.profile.get("edu.nrao.carta.path-to-executable")
    args.redis_host = args.profile.get("edu.nrao.carta.redis-host")
    args.redis_port = int(args.profile.get("edu.nrao.carta.redis-port"))
    args.redis_password = args.profile.get("edu.nrao.carta.redis-password")
    args.reverse_proxy_host = args.profile.get("edu.nrao.carta.reverse-proxy-host")
    return args


def perform_setup() -> SetupValues:
    """
    Perform necessary setup steps for the valet

    :return: Tuple containing the values needed to run CARTA in main
    """
    global CONN, REDIS_VALUES

    args = set_args(make_arg_parser().parse_args())
    front_end_id = generate_random_str()
    back_end_id = generate_random_str()
    front_end_port, back_end_port = get_available_port(), get_available_port()
    carta_url = f"https://{args.reverse_proxy_host}/{front_end_id}/?socketUrl=wss://{args.reverse_proxy_host}/{back_end_id}/"
    CONN = connect_to_redis(args.redis_host, args.redis_port, args.redis_password)
    REDIS_VALUES = get_redis_values(
        args.reverse_proxy_host,
        front_end_id,
        back_end_id,
        front_end_port,
        back_end_port,
    )
    while duplicate_redis_values_found(REDIS_VALUES, CONN):
        # Keep generating random IDs until there aren't any collisions in the Redis server
        logger.warning("WARNING: Redis value collision found. Generating new random IDs.")
        new_front_end_id = generate_random_str()
        new_back_end_id = generate_random_str()
        REDIS_VALUES = get_redis_values(
            args.reverse_proxy_host,
            new_front_end_id,
            new_back_end_id,
            front_end_port,
            back_end_port,
        )

    return args, front_end_port, back_end_port, carta_url


def make_arg_parser() -> argparse.ArgumentParser:
    """
    Create command-line argument parser

    :return: Argument parser
    """
    parser = argparse.ArgumentParser(
        "carta_valet",
        description=_DESCRIPTION,
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument(
        "-P",
        "--profile",
        action="store",
        type=str,
        required=False,
        help="name of CAPO profile to use",
    )
    parser.add_argument(
        "-t",
        "--timeout",
        action="store",
        type=int,
        required=False,
        help="number of minutes to keep CARTA open for",
    )
    parser.add_argument(
        "-w",
        "--where",
        action="store",
        type=str,
        required=False,
        help="where to run CARTA",
    )
    return parser


def main():
    """
    Entry point for CARTA valet
    """
    global REDIS_VALUES

    args, front_end_port, back_end_port, carta_url = perform_setup()
    # Run CARTA with setup/teardown
    setup_carta(REDIS_VALUES, args.timeout)
    run_carta(
        args.path_to_carta, args.timeout, args.where, front_end_port, back_end_port, carta_url
    )
    teardown_carta()


if __name__ == "__main__":
    main()
